%  Copyright (C) 2002-2004,2007 David Roundy
%  Copyright (C) 2005 Juliusz Chroboczek
%
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; see the file COPYING.  If not, write to
%  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
%  Boston, MA 02110-1301, USA.

\begin{code}
{-# OPTIONS -fglasgow-exts #-}
#include "gadts.h"
module Darcs.Repository.Internal ( Repository(..), RepoType(..), ($-),
                    maybeIdentifyRepository, identifyDarcs1Repository, identifyRepositoryFor,
                    findRepository, amInRepository, amNotInRepository,
                    slurp_pending, pristineFromWorking, revertRepositoryChanges,
                    slurp_recorded, slurp_recorded_and_unrecorded,
                    withRecorded, checkPristineAgainstCwd,
                    get_unrecorded, get_unrecorded_unsorted, get_unrecorded_no_look_for_adds,
                    read_repo, sync_repo,
                    unsafe_fast_read_repo,
                    prefsUrl,
                    read_pending, add_to_pending,
                    withRepoLock, withRepository, withRepositoryDirectory, withGutsOf,
                    tentativelyAddPatch, tentativelyRemovePatches, tentativelyAddToPending,
                    tentativelyReplacePatches,
                    tentativelyMergePatches, considerMergeToWorking,
                    finalizeRepositoryChanges,
                    unrevertUrl,
                    applyToWorking, patchSetToPatches,
                    createPristineDirectoryTree, createPartialsPristineDirectoryTree,
                    optimizeInventory,
                    getMarkedupFile,
                    PatchSet,
                  ) where
import Prelude hiding ( catch )

import Printer ( putDocLn, (<+>), text, ($$) )

import Data.Maybe ( isJust, isNothing )
import System.IO.Unsafe ( unsafeInterleaveIO )
import Darcs.Repository.Prefs ( get_prefval, unsafeCastCache )
import Darcs.Resolution ( standard_resolution, external_resolution )
import System ( ExitCode(..), exitWith, system )
import Darcs.External ( backupByCopying, fetchFilePS, Cachable( Cachable ),
                        clonePartialsTree )
import Darcs.IO ( runTolerantly, runSilently )
import Darcs.Repository.Pristine ( Pristine, identifyPristine, nopristine, checkPristine,
                                   easyCreatePristineDirectoryTree, slurpPristine, syncPristine,
                                   easyCreatePartialsPristineDirectoryTree,
                                   createPristineFromWorking )
import Data.List ( (\\) )
import Darcs.SignalHandler ( withSignalsBlocked )
import Darcs.Repository.Format ( RepoFormat, RepoProperty( Darcs2, HashedInventory ),
                                 identifyRepoFormat, format_has,
                                 write_problem, read_problem, readfrom_and_writeto_problem )
import System.Directory ( doesDirectoryExist, setCurrentDirectory, removeFile )
import Control.Monad ( liftM, when, unless )
import Workaround ( createDirectoryIfMissing, getCurrentDirectory, renameFile )
import FastPackedString ( PackedString, readFilePS, gzReadFilePS, nilPS, packString )
import Darcs.Patch ( is_hunk, is_binary, description,
                     try_to_shrink, commuteFL )
import Darcs.Patch.Prim ( try_shrinking_inverse )
import Darcs.Bug ( bugDoc )
import Darcs.Patch.Bundle ( scan_bundle, make_bundle )
import Darcs.SlurpDirectory ( Slurpy, slurp_unboring, mmap_slurp, co_slurp,
                              slurp_has )
import Darcs.Hopefully ( PatchInfoAnd, info,
                         hopefully, hopefullyM )
import Darcs.Repository.DarcsRepo ( lazily_read_repo )
import Darcs.Repository.ApplyPatches ( apply_patches )
import qualified Darcs.Repository.HashedRepo as HashedRepo
                            ( revert_tentative_changes, finalize_tentative_changes,
                              remove_from_tentative_inventory,
                              copy_pristine, copy_partials_pristine, slurp_pristine,
                              apply_to_tentative_pristine, pristine_from_working,
                              read_repo_lazily, write_tentative_inventory,
                              add_to_tentative_inventory, read_repo )
import qualified Darcs.Repository.DarcsRepo as DarcsRepo
import qualified Darcs.Repository.GitRepo as GitRepo
import Darcs.Flags ( DarcsFlag(AnyOrder, Boring, LookForAdds, Verbose, Quiet,
                               MarkConflicts, AllowConflicts, NoUpdateWorking,
                               RepoDir, WorkDir, UMask, Test, LeaveTestDir),
                     want_external_merge )
import Darcs.Patch.Ordered ( FL(..), RL(..),
                             (:\/:)(..), (:/\:)(..), (:>)(..),
                             (+>+), lengthFL,
                             unsafeFL, unsafeUnFL, allFL,
                             reverseRL, reverseFL, concatRL, mapFL,
                             mapFL_FL, concatFL )
import Darcs.Patch ( RepoPatch, Patchy, Prim, RealPatch, Patch, merge,
                     joinPatches, sort_coalesceFL,
                     list_conflicted_files, list_touched_files,
                     Named, patchcontents, anonymous,
                     commuteRL, fromPrims,
                     patch2patchinfo, readPatch,
                     writePatch, effect, invert,
                     is_addfile, is_adddir, is_addrmfile, is_addrmdir,
                     apply, apply_to_slurpy,
                     empty_markedup_file, MarkedUpFile
                   )
import Darcs.Patch.Info ( PatchInfo )
import Darcs.Patch.Set ( PatchSet )
import Darcs.Patch.Apply ( markup_file, LineMark(None) )
import Darcs.Patch.Depends ( get_common_and_uncommon )
import Darcs.Diff ( smart_diff )
import Darcs.Utils ( promptChar, catchall, withCurrentDirectory, withUMask,
                     prettyException, nubsort )
import Darcs.FilePathUtils ( absolute_dir )
import Darcs.URL ( is_file )
import Darcs.Repository.Prefs ( darcsdir_filter, boring_file_filter, filetype_function,
                                Cache, getCaches )
import Darcs.Lock ( withLock, writeDocBinFile, withDelayedDir, removeFileMayNotExist,
                    withTempDir, withPermDir )
import Control.Exception ( catch )
import Darcs.Sealed ( Sealed(Sealed), unsafeUnseal )

#include "impossible.h"

data Repository p = Repo !String ![DarcsFlag] !RepoFormat !(RepoType p) deriving ( Show )

data RepoType p = DarcsRepository !Pristine (Cache p) | GitRepository deriving ( Show )

maybeIdentifyRepository :: [DarcsFlag] -> String -> IO (Either String (Repository p))
maybeIdentifyRepository opts "." =
    do darcs <- doesDirectoryExist "_darcs"
       git <- doesDirectoryExist ".git"
       rf <- identifyRepoFormat "."
       here <- absolute_dir "."
       case read_problem rf of
         Just err -> return $ Left err
         Nothing ->
           case (darcs, git) of
           (False, False) -> return (Left "Not a repository")
           (True, _)->
               do pris <- identifyPristine
                  cs <- getCaches opts here
                  return $ Right $ Repo here opts rf (DarcsRepository pris cs)
           (False, True) -> return (Right $ Repo here opts rf GitRepository)
maybeIdentifyRepository opts url' = do
    url <- absolute_dir url'
    darcs <- doesRemoteFileExist (url++"/_darcs/inventory")
    hash_darcs <- doesRemoteFileExist (url++"/_darcs/hashed_inventory")
    git <- doesRemoteFileExist (url++"/.git/HEAD")
    rf <- identifyRepoFormat url
    case read_problem rf of
        Just err -> return $ Left err
        Nothing ->
            case (hash_darcs, darcs, git) of
            (Left _, Left s, Left _) ->
                return (Left $ "Not a repository: " ++ url' ++
                        " (" ++ s ++")")
            (Left _, Left _, Right True) -> return (Right $ Repo url opts rf GitRepository)
            _ -> do cs <- getCaches opts url
                    return $ Right $ Repo url opts rf (DarcsRepository nopristine cs)
    where drfe x = fetchFilePS x Cachable >> return True
          doesRemoteFileExist x = (liftM Right) (drfe x) `catch`
                                      (\e -> return (Left (prettyException e)))

identifyDarcs1Repository :: [DarcsFlag] -> String -> IO (Repository Patch)
identifyDarcs1Repository opts url =
    do er <- maybeIdentifyRepository opts url
       case er of
           Left s -> fail s
           Right r -> return r

identifyRepositoryFor :: RepoPatch p => Repository p -> String -> IO (Repository p)
identifyRepositoryFor (Repo _ opts rf _) url =
    do Repo absurl _ rf' t <- identifyDarcs1Repository opts url
       let t' = case t of DarcsRepository x c -> DarcsRepository x (unsafeCastCache c)
                          GitRepository -> GitRepository
       case readfrom_and_writeto_problem rf' rf of
         Just e -> fail $ "Incompatibility with repository " ++ url ++ ":\n" ++ e
         Nothing -> return $ Repo absurl opts rf' t'

isRight :: Either a b -> Bool
isRight (Right _) = True
isRight _         = False

currentDirIsRepository :: IO Bool
currentDirIsRepository = isRight `liftM` maybeIdentifyRepository [] "."

amInRepository :: [DarcsFlag] -> IO (Either String FilePath)
amInRepository (WorkDir d:_) =
    do setCurrentDirectory d `catchall` (fail $ "can't set directory to "++d)
       amInRepository []
amInRepository (_:fs) = amInRepository fs
amInRepository [] = 
    seekRepo (Left "You need to be in a repository directory to run this command.")

-- | hunt upwards for the darcs repository
-- This keeps changing up one parent directory, testing at each
-- step if the current directory is a repository or not.  $
-- WARNING this changes the current directory for good if matchFn succeeds
seekRepo :: Either String FilePath
            -- ^ what to return if we don't find a repository
         -> IO (Either String FilePath)
seekRepo onFail = getCurrentDirectory >>= helper "" where
   helper dir startpwd = do
    air <- currentDirIsRepository
    if air
       then return (Right dir)
       else do cd <- getCurrentDirectory
               setCurrentDirectory ".."
               cd' <- getCurrentDirectory
               if cd' /= cd
                  then helper (reverse (takeWhile (/='/') $ reverse cd)///dir) startpwd
                  else do setCurrentDirectory startpwd
                          return onFail
    where ""///b = b
          a///"" = a
          a///b = a ++ "/" ++ b


amNotInRepository :: [DarcsFlag] -> IO (Either String FilePath)
amNotInRepository (WorkDir d:_) = do createDirectoryIfMissing False d
                                     -- note that the above could always fail
                                     setCurrentDirectory d
                                     amNotInRepository []
amNotInRepository (_:f) = amNotInRepository f
amNotInRepository [] =
    do air <- currentDirIsRepository
       if air then return (Left $ "You may not run this command in a repository.")
              else return $ Right ""

findRepository :: [DarcsFlag] -> IO (Either String FilePath)
findRepository (RepoDir d:_) | is_file d =
    do setCurrentDirectory d `catchall` (fail $ "can't set directory to "++d)
       findRepository []
findRepository (WorkDir d:_) =
    do setCurrentDirectory d `catchall` (fail $ "can't set directory to "++d)
       findRepository []
findRepository (_:fs) = findRepository fs
findRepository [] = seekRepo (Right "")

slurp_pending :: RepoPatch p => Repository p -> IO Slurpy
slurp_pending repo = do
  cur <- slurp_recorded repo
  pend <- read_pending repo
  case apply_to_slurpy pend cur of
    Just pendcur -> return pendcur
    Nothing -> do putStrLn "Yikes, pending has conflicts!"
                  return cur

slurp_recorded :: RepoPatch p => Repository p -> IO Slurpy
slurp_recorded (Repo dir opts rf (DarcsRepository _ c))
    | format_has HashedInventory rf =
        HashedRepo.slurp_pristine c opts dir "_darcs/hashed_inventory"
slurp_recorded repository@(Repo dir _ _ (DarcsRepository p _)) = do
    mc <- withCurrentDirectory dir $ slurpPristine p
    case mc of (Just slurpy) -> return slurpy
               Nothing -> withDelayedDir "pristine.temp" $ \cd ->
                          do createPristineDirectoryTree repository cd
                             mmap_slurp cd
slurp_recorded (Repo r _ _ GitRepository) = GitRepo.slurpHead r

slurp_recorded_and_unrecorded :: RepoPatch p => Repository p -> IO (Slurpy, Slurpy)
slurp_recorded_and_unrecorded repo@(Repo r _ _ _) = do
  cur <- slurp_recorded repo
  pend <- read_pending repo
  withCurrentDirectory r $
      case apply_to_slurpy pend cur of
      Nothing -> fail "Yikes, pending has conflicts!"
      Just pendslurp -> do unrec <- co_slurp pendslurp "."
                           return (cur, unrec)

pendingName :: RepoType p -> String
pendingName (DarcsRepository _ _) = "_darcs/patches/pending"
pendingName GitRepository = ".git/darcs_pending"

read_pending :: RepoPatch p => Repository p -> IO (FL Prim)
read_pending (Repo r _ _ tp) =
    withCurrentDirectory r (read_pendingfile (pendingName tp))

add_to_pending :: RepoPatch p => Repository p -> FL Prim -> IO ()
add_to_pending (Repo _ opts _ _) _ | NoUpdateWorking `elem` opts = return ()
add_to_pending repo p =
    do pend <- get_unrecorded repo
       make_new_pending repo (pend +>+ p)

readPrims :: PackedString -> Sealed (FL Prim C(x))
readPrims s = case readPatch s of
              Nothing -> Sealed NilFL
              Just (Sealed p,_) -> Sealed (effect (p :: Patch C(x y)))

read_pendingfile :: String -> IO (FL Prim)
read_pendingfile name = do
  pend <- gzReadFilePS name `catch`
          (\_ -> return $ packString "")
  case readPrims pend of
    Sealed p -> return p

make_new_pending :: RepoPatch p => Repository p -> FL Prim -> IO ()
make_new_pending (Repo _ opts _ _) _ | NoUpdateWorking `elem` opts = return ()
make_new_pending repo@(Repo r _ _ tp) origp =
    withCurrentDirectory r $
    do let newname = pendingName tp ++ ".new"
       writePatch newname (fromPrims $ sift_for_pending origp :: Patch)
       cur <- slurp_recorded repo
       p <- read_pendingfile newname
       when (isNothing $ apply_to_slurpy p cur) $ do
         let buggyname = pendingName tp ++ "_buggy"
         renameFile newname buggyname
         bugDoc $ text "There was an attempt to write an invalid pending!"
                    $$ text "If possible, please send the contents of"
                    <+> text buggyname
                    $$ text "along with a bug report."
       renameFile newname (pendingName tp)

sift_for_pending :: FL Prim -> FL Prim
sift_for_pending simple_ps =
 let oldps = maybe simple_ps id $ try_shrinking_inverse simple_ps
 in if allFL (\p -> is_addfile p || is_adddir p) $ oldps
    then oldps
    else 
      case try_to_shrink $ sfp NilFL $ reverseFL oldps of
      ps | lengthFL ps < lengthFL oldps -> sift_for_pending ps
         | otherwise -> ps
      where sfp :: FL Prim C(x y) -> RL Prim C(y z) -> FL Prim C(x z)
            sfp sofar NilRL = sofar
            sfp sofar (p:<:ps)
                | is_hunk p || is_binary p
                    = case commuteFL (p :> sofar) of
                      Just (sofar' :> _) -> sfp sofar' ps
                      Nothing -> sfp (p:>:sofar) ps
            sfp sofar (p:<:ps) = sfp (p:>:sofar) ps

get_unrecorded_no_look_for_adds :: RepoPatch p => Repository p -> IO (FL Prim)
get_unrecorded_no_look_for_adds = get_unrecorded_private (filter (/= LookForAdds))

get_unrecorded_unsorted :: RepoPatch p => Repository p -> IO (FL Prim)
get_unrecorded_unsorted = get_unrecorded_private (AnyOrder:)

get_unrecorded :: RepoPatch p => Repository p -> IO (FL Prim)
get_unrecorded = get_unrecorded_private id

get_unrecorded_private :: RepoPatch p => ([DarcsFlag]->[DarcsFlag]) -> Repository p -> IO (FL Prim)
get_unrecorded_private _ (Repo _ opts _ _)
    | NoUpdateWorking `elem` opts = return NilFL
get_unrecorded_private modopts repository@(Repo r oldopts _ _) =
  withCurrentDirectory r $ do
    cur <- slurp_pending repository
    work <- if LookForAdds `elem` opts
            then do nboring <- if Boring `elem` opts
                               then return $ darcsdir_filter
                               else boring_file_filter
                    slurp_unboring (myfilt cur nboring) "."
            else co_slurp cur "."
    pend <- read_pending repository
    when (Verbose `elem` opts) $ putStrLn "diffing dir..."
    ftf <- filetype_function
    case smart_diff opts ftf cur work of
      di -> if AnyOrder `elem` opts
            then return $ pend +>+ di
            else return $ sort_coalesceFL $ pend +>+ di
    where myfilt s nboring f = slurp_has f s || nboring [f] /= []
          opts = modopts oldopts

read_repo :: RepoPatch p => Repository p -> IO (PatchSet p)
read_repo (Repo r opts rf (DarcsRepository _ c))
    | format_has HashedInventory rf = HashedRepo.read_repo c opts r
    | otherwise = DarcsRepo.read_repo opts r
read_repo (Repo r _ _ GitRepository) = GitRepo.read_repo r

-- unsafe_fast_read_repo parses the patches lazily, assuming that they will
-- both exist and be valid files.  This is obviously unsafe in the senses
-- that "isJust" doesn't tell us if a patch is actually parsable.
unsafe_fast_read_repo :: RepoPatch p => Repository p -> IO (PatchSet p)
unsafe_fast_read_repo (Repo r opts rf (DarcsRepository _ c))
    | format_has HashedInventory rf = HashedRepo.read_repo_lazily c opts r
    | otherwise = DarcsRepo.lazily_read_repo opts r
unsafe_fast_read_repo (Repo r _ _ GitRepository) = GitRepo.read_repo r

sync_repo :: Repository p -> IO ()
sync_repo (Repo r _ _ (DarcsRepository p _)) = withCurrentDirectory r $ syncPristine p
-- this should probably update the Git cache
sync_repo (Repo _ _ _ GitRepository) = return ()

prefsUrl :: Repository p -> String
prefsUrl (Repo r _ _ (DarcsRepository _ _)) = r ++ "/_darcs/prefs"
prefsUrl (Repo r _ _ GitRepository) = r ++ "/.git/darcs-prefs"

unrevertUrl :: Repository p -> String
unrevertUrl (Repo r _ _ (DarcsRepository _ _)) = r ++ "/_darcs/patches/unrevert"
unrevertUrl (Repo r _ _ GitRepository) = r ++ "/.git/darcs-unrevert"

applyToWorking :: Patchy p => Repository p1 -> [DarcsFlag] -> p -> IO ()
applyToWorking (Repo r _ _ (DarcsRepository _ _)) opts patch =
    withCurrentDirectory r $ if Quiet `elem` opts
                             then runSilently $ apply opts patch
                             else runTolerantly $ apply opts patch
applyToWorking (Repo _ _ _ GitRepository) _ _ = return ()

handle_pend_for_add :: RepoPatch p => Repository p -> [DarcsFlag] -> (Named p) -> IO ()
handle_pend_for_add (Repo _ opts _ _) _ _ | NoUpdateWorking `elem` opts = return ()
handle_pend_for_add repository opts p =
    do pend <- read_pending repository
       let is_okay pa | is_addrmfile pa || is_addrmdir pa =
                          head (list_touched_files pa) `elem`
                          list_touched_files pend
                      | otherwise = True
           filter_p :: [Prim] -> [Prim]
           filter_p = if LookForAdds `elem` opts
                      then filter is_okay
                      else id
       prepend repository $ invert $ unsafeFL $ filter_p $ unsafeUnFL $ np2prims p

data HashedVsOld a = HvsO { old, hashed :: a }

decideHashedOrNormal :: Monad m => RepoFormat -> HashedVsOld (m a) -> m a
decideHashedOrNormal rf (HvsO { hashed = h, old = o })
    | format_has HashedInventory rf = h
    | otherwise = o

tentativelyMergePatches :: RepoPatch p => Repository p -> String -> [DarcsFlag]
                        -> FL (Named p) -> FL (Named p) -> IO (FL Prim)
tentativelyMergePatches = tentativelyMergePatches_ MakeChanges

considerMergeToWorking :: RepoPatch p => Repository p -> String -> [DarcsFlag]
                       -> FL (Named p) -> FL (Named p) -> IO (FL Prim)
considerMergeToWorking = tentativelyMergePatches_ DontMakeChanges

data MakeChanges = MakeChanges | DontMakeChanges deriving ( Eq )

tentativelyMergePatches_ :: RepoPatch p => MakeChanges
                         -> Repository p -> String -> [DarcsFlag] -> FL (Named p) -> FL (Named p) -> IO (FL Prim)
tentativelyMergePatches_ mc r cmd opts us them =
  do let pc = case merge (them :\/: us) of
              _ :/\: x -> x
         putVerbose = if Verbose `elem` opts then putStrLn else \_ -> return ()
     pend <- get_unrecorded_unsorted r -- we don't care if it looks pretty...
     let pw = case merge (pc :\/: anonymous (fromPrims pend) :>: NilFL) of
                _ :/\: p' -> p'
         pwprim = joinPatches $ mapFL_FL patchcontents pw
     let Sealed standard_resolved_pw = standard_resolution pwprim
     mapM_ backupByCopying $ list_touched_files standard_resolved_pw
     have_conflicts <- announce_merge_conflicts cmd opts standard_resolved_pw
     have_unrecorded_conflicts <- check_unrecorded_conflicts opts pc
     (_, working) <- slurp_recorded_and_unrecorded r
     pw_resolution <-
          case (want_external_merge opts, have_conflicts || have_unrecorded_conflicts) of
          (Nothing,_) -> return $ if AllowConflicts `elem` opts
                                  then NilFL
                                  else standard_resolved_pw
          (_,False) -> return standard_resolved_pw
          (Just c, True) -> unsafeUnseal `fmap`
                            external_resolution working c
                                                    (nps2prims us +>+ pend)
                                                    (nps2prims them) pwprim
     putVerbose "Applying patches to the local directories..."
     when (mc == MakeChanges) $
          do sequence_ $ mapFL (tentativelyAddPatch r opts) pc
             tentativelyAddToPending r opts (nps2prims pw +>+ pw_resolution)
     return (nps2prims pw +>+ pw_resolution)

announce_merge_conflicts :: String -> [DarcsFlag] -> FL Prim -> IO Bool
announce_merge_conflicts cmd opts resolved_pw =
    case nubsort $ list_touched_files $ resolved_pw of
    [] -> return False
    cfs -> if MarkConflicts `elem` opts || AllowConflicts `elem` opts
              || want_external_merge opts /= Nothing
           then do putStrLn "We have conflicts in the following files:"
                   putStrLn $ unwords cfs
                   return True
           else do putStrLn "There are conflicts in the following files:"
                   putStrLn $ unwords cfs
                   fail $ "Refusing to "++cmd++" patches leading to conflicts.\n"++
                          "If you would rather apply the patch and mark the conflicts,\n"++
                          "use the --mark-conflicts option to "++cmd++"."

np2prims :: RepoPatch p => Named p C(x y) -> FL Prim C(x y)
np2prims = effect . patchcontents

nps2prims :: RepoPatch p => FL (Named p) C(x y) -> FL Prim C(x y)
nps2prims = concatFL . mapFL_FL np2prims

check_unrecorded_conflicts :: forall p. RepoPatch p => [DarcsFlag] -> FL (Named p) -> IO Bool
check_unrecorded_conflicts opts _ | NoUpdateWorking `elem` opts = return False
check_unrecorded_conflicts opts pc =
    do repository <- identifyDarcs1Repository opts "."
       mpend <- read_pending repository
       case mpend of
         NilFL -> return False
         pend ->
             case merge (fromPrims pend :\/: fromPrims (concatFL $ mapFL_FL effect pc)) of
             _ :/\: (pend' :: p) ->
                 case list_conflicted_files pend' of
                 [] -> return False
                 fs -> do yorn <- promptChar
                                  ("You have conflicting local changes to:\n"
                                   ++ unwords fs++"\nProceed?") "yn"
                          when (yorn /= 'y') $
                               do putStr "Cancelled."
                                  exitWith ExitSuccess
                          return True

tentativelyAddPatch :: RepoPatch p => Repository p -> [DarcsFlag] -> (Named p) -> IO (Named p)
tentativelyAddPatch = tentativelyAddPatch_ UpdatePristine

data UpdatePristine = UpdatePristine | DontUpdatePristine deriving Eq

tentativelyAddPatch_ :: RepoPatch p => UpdatePristine -> Repository p -> [DarcsFlag]
                     -> (Named p) -> IO (Named p)
tentativelyAddPatch_ up r@(Repo dir _ rf (DarcsRepository _ c)) opts patch =
    withCurrentDirectory dir $
    do fp <- decideHashedOrNormal rf $ HvsO {
          hashed = HashedRepo.add_to_tentative_inventory c patch,
          old = DarcsRepo.add_to_tentative_inventory opts patch
                                         }
       when (up == UpdatePristine) $
            do when (Verbose `elem` opts) $ putDocLn $ text "Applying to pristine..." <+> description patch
               decideHashedOrNormal rf $ HvsO {hashed = HashedRepo.apply_to_tentative_pristine c opts patch,
                                               old = DarcsRepo.add_to_tentative_pristine patch}
               handle_pend_for_add r opts patch
       unsafeInterleaveIO $ do ps <- gzReadFilePS (dir++"/"++fp)
                               case readPatch ps of
                                 Just (Sealed x,_) -> return x
                                 _ -> fail $ "Error rereading patch " ++ dir++"/"++fp
tentativelyAddPatch_ up r@(Repo dir _ _ GitRepository) opts patch =
    withCurrentDirectory dir $
    do (patch', token) <- GitRepo.writePatch dir patch
       when (up == UpdatePristine) $ do handle_pend_for_add r opts patch'
                                        GitRepo.set_tentative_tree token
       return patch'

tentativelyAddToPending :: forall p. RepoPatch p => Repository p -> [DarcsFlag] -> FL Prim -> IO ()
tentativelyAddToPending (Repo _ opts _ _) _ _ | NoUpdateWorking `elem` opts = return ()
tentativelyAddToPending (Repo dir _ _ rt) _ patch =
    withCurrentDirectory dir $ do
      let pn = pendingName rt
          tpn = pn ++ ".tentative"
      pend <- gzReadFilePS tpn `catchall` (return nilPS)
      let newpend = case readPrims pend of
                    Sealed NilFL -> patch
                    Sealed p -> p +>+ patch
      writePatch tpn $ (fromPrims newpend :: Patch)

prepend :: forall p. RepoPatch p => Repository p -> FL Prim -> IO ()
prepend (Repo _ opts _ _) _ | NoUpdateWorking `elem` opts = return ()
prepend (Repo _ _ _ rt) patch = do let ppn = pendingName rt ++ ".tentative_prepend"
                                   pend <- gzReadFilePS ppn `catchall` (return nilPS)
                                   let newpend = case readPrims pend of
                                                 Sealed NilFL -> patch
                                                 Sealed p -> patch +>+ p
                                   writePatch ppn $ (fromPrims newpend :: Patch)

tentativelyRemovePatches :: RepoPatch p => Repository p -> [DarcsFlag]
                         -> FL (Named p) -> IO ()
tentativelyRemovePatches = tentativelyRemovePatches_ UpdatePristine

tentativelyRemovePatches_ :: RepoPatch p => UpdatePristine
                          -> Repository p -> [DarcsFlag]
                          -> FL (Named p) -> IO ()
tentativelyRemovePatches_ up repository@(Repo dir _ rf (DarcsRepository _ c)) opts ps =
    withCurrentDirectory dir $ do
      when (up == UpdatePristine) $
           prepend repository $ nps2prims ps
      remove_from_unrevert_context repository ps
      if format_has HashedInventory rf
        then do HashedRepo.remove_from_tentative_inventory c opts ps
                when (up == UpdatePristine) $ HashedRepo.apply_to_tentative_pristine c opts $ invert ps
        else DarcsRepo.remove_from_tentative_inventory updarcspris opts ps
               where updarcspris = up==UpdatePristine && not (format_has HashedInventory rf)
tentativelyRemovePatches_ _ (Repo _ _ _ GitRepository) _ _ =
    fail "unimplemented git"

tentativelyReplacePatches :: RepoPatch p => Repository p -> [DarcsFlag]
                          -> FL (Named p) -> IO ()
tentativelyReplacePatches repository opts ps =
    do tentativelyRemovePatches_ DontUpdatePristine repository opts ps
       sequence_ $ mapFL
           (tentativelyAddPatch_ DontUpdatePristine repository opts) ps

finalize_pending :: RepoPatch p => Repository p -> IO ()
finalize_pending (Repo dir opts _ rt)
    | NoUpdateWorking `elem` opts =
        withCurrentDirectory dir $ removeFileMayNotExist $ (pendingName rt)
finalize_pending repository@(Repo dir _ _ rt) = do
  withCurrentDirectory dir $ do let pn = pendingName rt
                                    tpn = pn ++ ".tentative"
                                    ppn = pn ++ ".tentative_prepend"
                                pend <- read_pending repository
                                ppfile <- gzReadFilePS ppn `catchall` (return nilPS)
                                tpfile <- gzReadFilePS tpn `catchall` (return nilPS)
                                let tpend = unsafeUnseal $ readPrims tpfile
                                    ppend = unsafeUnseal $ readPrims ppfile
                                    new_pending = sift_for_pending $ ppend+>+pend+>+tpend
                                make_new_pending repository new_pending

finalizeRepositoryChanges :: RepoPatch p => Repository p -> IO ()
finalizeRepositoryChanges repository@(Repo dir opts rf (DarcsRepository _ c))
    | format_has HashedInventory rf =
        withCurrentDirectory dir $ do testTentative repository
                                      withSignalsBlocked $ do HashedRepo.finalize_tentative_changes c opts
                                                              finalize_pending repository
finalizeRepositoryChanges repository@(Repo dir _ _ (DarcsRepository _ _)) =
  withCurrentDirectory dir $ do testTentative repository
                                withSignalsBlocked $ do DarcsRepo.finalize_pristine_changes
                                                        DarcsRepo.finalize_tentative_changes
                                                        finalize_pending repository
finalizeRepositoryChanges repository@(Repo dir _ _ GitRepository) =
    withCurrentDirectory dir $ do GitRepo.finalize_tentative_changes
                                  finalize_pending repository

testTentative :: RepoPatch p => Repository p -> IO ()
testTentative repository@(Repo dir opts _ _) =
    when (Test `elem` opts) $ withCurrentDirectory dir $
    do let putInfo = if not $ Quiet `elem` opts then putStrLn else const (return ())
       testline <- get_prefval "test"
       case testline of
         Nothing -> return ()
         Just testcode ->
             withTentative repository (wd "testing") $ \_ ->
             do putInfo "Running test...\n"
                ec <- system testcode
                if ec == ExitSuccess
                  then putInfo "Test ran successfully.\n"
                  else do putInfo "Test failed!\n"
                          exitWith ec
    where wd = if LeaveTestDir `elem` opts then withPermDir else withTempDir

revertRepositoryChanges :: Repository p -> IO ()
revertRepositoryChanges (Repo dir opts rf dr@(DarcsRepository _ _)) =
    withCurrentDirectory dir $
    do removeFileMayNotExist $ (pendingName dr) ++ ".tentative"
       removeFileMayNotExist $ (pendingName dr) ++ ".tentative_prepend"
       when (NoUpdateWorking `elem` opts) $ removeFileMayNotExist $ pendingName dr
       decideHashedOrNormal rf $ HvsO { hashed = HashedRepo.revert_tentative_changes,
                                        old = DarcsRepo.revert_tentative_changes }
revertRepositoryChanges (Repo dir _ _ GitRepository) =
    withCurrentDirectory dir $
    do removeFileMayNotExist $ (pendingName GitRepository) ++ ".tentative"
       removeFileMayNotExist $ (pendingName GitRepository) ++ ".tentative_prepend"
       GitRepo.revert_tentative_changes

patchSetToPatches :: RepoPatch p => PatchSet p -> FL (Named p)
patchSetToPatches patchSet = mapFL_FL hopefully $ reverseRL $ concatRL patchSet

getUMask :: [DarcsFlag] -> Maybe String
getUMask [] = Nothing
getUMask ((UMask u):_) = Just u
getUMask (_:l) = getUMask l

withGutsOf :: Repository p -> IO () -> IO ()
withGutsOf (Repo _ _ rf _) | format_has HashedInventory rf = id
                           | otherwise = withSignalsBlocked

withRepository :: [DarcsFlag] -> (forall p. RepoPatch p => Repository p -> IO a) -> IO a
withRepository opts1 = withRepositoryDirectory opts1 "."

withRepositoryDirectory :: [DarcsFlag] -> String
                        -> (forall p. RepoPatch p => Repository p -> IO a) -> IO a
withRepositoryDirectory opts1 url job =
    do Repo dir opts rf rt <- identifyDarcs1Repository opts1 url
       let rt' = case rt of DarcsRepository t c -> DarcsRepository t (unsafeCastCache c)
                            GitRepository -> GitRepository
       if format_has Darcs2 rf
         then job (Repo dir opts rf rt' :: Repository (FL RealPatch))
         else job (Repo dir opts rf rt :: Repository Patch)

($-) :: ((forall p. RepoPatch p => Repository p -> IO a) -> IO a)
     -> (forall p. RepoPatch p => Repository p -> IO a) -> IO a
x $- y = x y

withRepoLock :: [DarcsFlag] -> (forall p. RepoPatch p => Repository p -> IO a) -> IO a
withRepoLock opts job =
    withRepository opts $- \repository@(Repo _ _ rf rt) ->
    do case write_problem rf of
         Nothing -> return ()
         Just err -> fail err
       let name = case rt of
                    (DarcsRepository _ _) -> "./_darcs/lock"
                    GitRepository -> "./.git/lock"
           wu = case (getUMask opts) of
                  Nothing -> id
                  Just u -> withUMask u
       wu $ withLock name
              (revertRepositoryChanges repository >> job repository)
\end{code}

\begin{code}
remove_from_unrevert_context :: RepoPatch p => Repository p -> FL (Named p) -> IO ()
remove_from_unrevert_context repository ps = do
  bundle <- unrevert_patch_bundle `catchall` return (NilRL:<:NilRL)
  case bundle of
    NilRL:<:NilRL -> return ()
    _ -> do
    let unrevert_loc = unrevertUrl repository
    ref <- read_repo repository
    case get_common_and_uncommon (bundle, ref) of
        (common,(h_us:<:NilRL):<:NilRL,NilRL:<:NilRL) ->
            case commuteRL (reverseFL ps :> hopefully h_us) of
            Nothing -> unrevert_impossible unrevert_loc
            Just (us' :> _) -> do
                s <- slurp_recorded repository
                writeDocBinFile unrevert_loc $
                             make_bundle [] s
                             (common \\ pis) (us':>:NilFL)
        (common,(x:<:NilRL):<:NilRL,_)
            | isr && any (`elem` common) pis -> unrevert_impossible unrevert_loc
            | isr -> return ()
            where isr = isJust $ hopefullyM x
        _ -> unrevert_impossible unrevert_loc
  where unrevert_impossible unrevert_loc =
            do yorn <- promptChar
                       "This operation will make unrevert impossible!\nProceed?"
                       "yn"
               case yorn of
                 'n' -> fail "Cancelled to avoid unrevert catastrophe!"
                 'y' -> removeFile unrevert_loc `catchall` return ()
                 _ -> impossible
        pis = mapFL patch2patchinfo ps
        unrevert_patch_bundle :: RepoPatch p => IO (PatchSet p)
        unrevert_patch_bundle = do pf <- readFilePS (unrevertUrl repository)
                                   case scan_bundle pf of
                                     Right foo -> return foo
                                     Left err -> fail $ "Couldn't parse unrevert patch:\n" ++ err
\end{code}

\begin{code}
optimizeInventory :: RepoPatch p => Repository p -> IO ()
optimizeInventory repository@(Repo r _ rf (DarcsRepository _ c)) =
    do ps <- read_repo repository
       decideHashedOrNormal rf $
           HvsO { hashed = do revertRepositoryChanges repository
                              HashedRepo.write_tentative_inventory c ps
                              finalizeRepositoryChanges repository,
                  old = DarcsRepo.write_inventory r ps
                }
optimizeInventory (Repo _ _ _ GitRepository) = return ()

createPristineDirectoryTree :: RepoPatch p => Repository p -> FilePath -> IO ()
createPristineDirectoryTree repo@(Repo r opts rf (DarcsRepository pris c)) reldir
    | format_has HashedInventory rf =
        do createDirectoryIfMissing True reldir
           withCurrentDirectory reldir $ HashedRepo.copy_pristine c opts r "_darcs/hashed_inventory"
    | otherwise =
        do dir <- absolute_dir reldir
           done <- withCurrentDirectory r $ easyCreatePristineDirectoryTree pris dir
           unless done $ do patches <- (reverseRL . concatRL) `fmap` read_repo repo
                            createDirectoryIfMissing True dir
                            withCurrentDirectory dir $ apply_patches [] noPut noPut patches
                                where noPut _ = return ()

createPristineDirectoryTree repository@(Repo r _ _ _) reldir = do
  dir <- absolute_dir reldir
  withCurrentDirectory r $
    do pris <- identifyPristine
       done <- easyCreatePristineDirectoryTree pris dir
       unless done $ do
          patches <- (reverseRL . concatRL) `fmap` read_repo repository
          createDirectoryIfMissing True dir
          withCurrentDirectory dir $ apply_patches [] noPut noPut patches
                  where noPut _ = return ()

createPartialsPristineDirectoryTree :: RepoPatch p => Repository p -> [FilePath] -> FilePath -> IO ()
createPartialsPristineDirectoryTree (Repo r opts rf (DarcsRepository _ c)) prefs dir
    | format_has HashedInventory rf =
        do createDirectoryIfMissing True dir
           withCurrentDirectory dir $
               HashedRepo.copy_partials_pristine c opts r "_darcs/hashed_inventory" prefs
createPartialsPristineDirectoryTree r@(Repo rdir _ _ (DarcsRepository pris _)) prefs dir
 = withCurrentDirectory rdir $
   do done <- easyCreatePartialsPristineDirectoryTree prefs pris dir
      unless done $ withRecorded r (withTempDir "recorded") $ \_ -> do
          clonePartialsTree "." dir prefs
createPartialsPristineDirectoryTree _ _ _ = fail "GitRepository is not supported"

pristineFromWorking :: RepoPatch p => Repository p -> IO ()
pristineFromWorking (Repo dir opts rf (DarcsRepository _ c))
    | format_has HashedInventory rf =
        withCurrentDirectory dir $ HashedRepo.pristine_from_working c opts
pristineFromWorking (Repo dir _ _ (DarcsRepository p _)) =
  withCurrentDirectory dir $ createPristineFromWorking p
pristineFromWorking (Repo _ _ _ GitRepository) = bug "GitRepository not implemented"

withRecorded :: RepoPatch p => Repository p -> ((FilePath -> IO a) -> IO a) -> (FilePath -> IO a) -> IO a
withRecorded repository mk_dir f
    = mk_dir $ \d -> do createPristineDirectoryTree repository d
                        f d

checkPristineAgainstCwd :: RepoPatch p => Repository p -> IO Bool
checkPristineAgainstCwd (Repo dir _ rf (DarcsRepository p _))
    | not $ format_has HashedInventory rf = do here <- absolute_dir "."
                                               withCurrentDirectory dir $ checkPristine here p
checkPristineAgainstCwd repository@(Repo _ opts _ _) =
    do s2 <- mmap_slurp "."
       s1 <- slurp_recorded repository
       ftf <- filetype_function
       case smart_diff opts ftf s1 s2 of
         NilFL -> return True
         _ -> return False

withTentative :: forall p a. RepoPatch p =>
                 Repository p -> ((FilePath -> IO a) -> IO a) -> (FilePath -> IO a) -> IO a
withTentative repository@(Repo dir opts rf (DarcsRepository _ c)) mk_dir f
    | format_has HashedInventory rf =
        withRecorded repository mk_dir $ \d ->
        do HashedRepo.copy_pristine c opts dir "_darcs/tentative_pristine"
           f d
withTentative repository@(Repo dir opts _ _) mk_dir f =
    withRecorded repository mk_dir $ \d ->
    do ps <- read_patches (dir ++ "/_darcs/tentative_pristine")
       apply opts $ joinPatches ps
       f d
    where read_patches :: FilePath -> IO (FL p)
          read_patches fil = do ps <- readFilePS fil
                                return $ case readPatch ps of
                                           Just (Sealed x, _) -> x
                                           Nothing -> NilFL
\end{code}

\begin{code}
getMarkedupFile :: RepoPatch p => Repository p -> PatchInfo -> FilePath -> IO MarkedUpFile
getMarkedupFile repository pinfo f = do
  patches <- liftM (dropWhile ((/= pinfo) . info) . unsafeUnFL
                    . reverseRL . concatRL) $ read_repo repository
  return $ snd $ do_mark_all patches (f, empty_markedup_file)
do_mark_all :: RepoPatch p => [PatchInfoAnd p]
            -> (FilePath, MarkedUpFile) -> (FilePath, MarkedUpFile)
do_mark_all (hp:pps) (f, mk) =
    case hopefullyM hp of
    Just p -> do_mark_all pps $ markup_file (info hp) (patchcontents p) (f, mk)
    Nothing -> (f, [(packString "Error reading a patch!",None)])
do_mark_all [] (f, mk) = (f, mk)
\end{code}
